!!!!!!
!RUBC_wrapper.f90
!
!This program solves and simulates the baseline GE version of the RUBC model.
!The file is in wrapper form, implementing the SMM point estimation.
!
!"Really Uncertain Business Cycles"
!Nick Bloom, Max Floetotto, Nir Jaimovich, Itay Saporta-Eksten, Stephen J. Terry
!
!This Version: Sep 4, 2014
!!!!!!
module fixedparam
implicit none


double precision, parameter :: alpha = 0.25,&
    nu = 0.5,&
    theta = 2.0,&
    deltak = 0.026,&
    deltan = 0.088,&
    beta = 0.95 ** 0.25,&
    capirrev = 0.339,&
    capfix = 0.0,&
    labfix = 0.024*4,&
    hirelin = 0.018*4.0,&
    firelin = 0.018*4.0,&
    rhoz = 0.95,&
    rhoa = 0.95,&
    kmin = 1.125,&
    lmin = 0.035,&
    kbarmin = 3.0,&
    kbarmax = 10.0,&
    vferrortol = 1e-4,&
    perrortol = 0.5e-2,&
    plb = 0.01,&
    pub = 3.0,&
    disttol = 1e-4,&
    pwindow = 0.025,&
    GEupdate = 0.33,&
    fcsterrortol = 0.0125,&
    RMSEchangetol = 0.001,&
    R2changetol = 0.01,&
    maxDenHaanchangetol = 0.01,&
    avgDenHaanchangetol = 0.01,&
    changetol = 1e-10
    
integer, parameter :: checkbounds = 1,&
    GEerrorswitch = 4,&
    znum = 5,&
    anum = 5,&
    snum = 2,&
    knum = 91,&
    lnum = 37,&
    kbarnum = 10,&
    numper = 5000,&
    numdiscard = 500,&
    ainit = 3,&
    sinit = 1,&
    zinit = 3,&
    nfirms = 1000,&
    waitsec = 15,&
    vfmaxit=50,&
    accelmaxit = 200,&
    maxpit = 25,&
    pcutoff = 15,&
    maxGEit = 10,&
    numconstants = 14,&
    maxwaits=30

end module fixedparam

module macrodata
implicit none
double precision, parameter, dimension(4) :: macrodatamom = (/3.3559317 , 0.7556358 , 0.8259057 , 0.8787079/)
end module macrodata

module microdata
implicit none
double precision, parameter, dimension(4) :: microdatamom = (/39.2774376, 4.8941042 , 1.1627580 , 0.7465530/)
end module microdata

program RUBC_wrapper
use base_lib
use omp_lib
use fixedparam
use macrodata
use microdata  
implicit none

double precision :: SMMobjective,xpsotol,fpsotol,xquicktol,phipso(2),fpso

double precision, allocatable :: x(:),lb(:),ub(:),xpso(:),asimshocks(:),&
    ssimshocks(:),zfirmsshocks(:,:)

integer :: npart,xquicknum,maxpsoit,nparam,seeddim,seedint,ct,firmct,t,&
    psoseed,newrun

integer, allocatable :: seedarray(:)

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!This point starts the main program performing the minimization within the 
!structural estimation.

nparam = 6; !number of SMM estimation parameters

!call the random numbers outside the pso optimization, which has its own random seed
seedint = 2501; !random number seed
call random_seed(size=seeddim)

allocate(x(nparam),lb(nparam),ub(nparam),xpso(nparam),seedarray(seeddim),asimshocks(numper),&
    ssimshocks(numper),zfirmsshocks(nfirms,numper))

!seed the random numbers
do ct=1,seeddim
    seedarray(ct) = seedint+ct
end do !ct
call random_seed(put=seedarray)

!draw the random numbers
call random_number(asimshocks); !U(0,1) in each cell of asimshocks
call random_number(ssimshocks); !U(0,1) in each cell of ssimshocks
call random_number(zfirmsshocks); !U(0,1) in each entry of zfirmsshocks

!write the random numbers
open(8,file="asimshocks.txt")
do ct=1,numper
write(8,*) asimshocks(ct)
end do 
close(8)

open(8,file="ssimshocks.txt")
do ct=1,numper
write(8,*) ssimshocks(ct)
end do 
close(8)

open(8,file="zfirmsshocks.txt")
do firmct=1,nfirms
do t=1,numper
write(8,*) zfirmsshocks(firmct,t)
end do 
end do 
close(8)

!pso parameters
npart = 25
xpsotol = 1.0e-3
fpsotol = 1.0e-3
xquicktol = 1.0e-3
xquicknum = 5
phipso = (/2.05,2.05/)
maxpsoit = 5000

!parameter initial guess (note that this isn't a direct input into pso, simply initialization and bounds setting)
x(1) = 1.6060678164419682; !ajump
x(2) = 4.1365844523734641; !zjump
x(3) = 0.94296716663676217; !uncpers
x(4) = 0.025857201385718384; !uncfreq
x(5) = 0.050752106584379765; !sigmaz
x(6) = 0.0066771142154949305; !sigmaa

!do PSO optimization
xpso(:) = 0.0; fpso = 0.0; !initialize to input and output optimization parameters to 0

!set bounds for parameter optimizations
lb(:) = 0.0; ub(:) = 0.0;
lb(1) = x(1)*0.995; ub(1) = x(1)*1.005; !ajump - (10,75)%ile
lb(2) = x(2)*0.995; ub(2) = x(2)*1.005; !zjump - (25,90)%ile
lb(3) = x(3)*0.995; ub(3) = x(3)*1.005; !uncpers - (25,75)%ile
lb(4) = x(4)*0.995; ub(4) = x(4)*1.005; !uncfreq - (25,75)%ile
lb(5) = x(5)*0.995; ub(5) = x(5)*1.005; !sigmaz - (25,90)%ile
lb(6) = x(6)*0.995; ub(6) = x(6)*1.005; !sigmaa - (25,75)%ile

open(8,file="lb_ub.txt")
do ct=1,6
    write(8,*) lb(ct),ub(ct)
end do !ct
close(8)

!initialize the PSO optimization
psoseed = 2603
newrun = 0

!actually perform the SMM minimization
call psorestart(xpso,fpso,RUBC_func,lb,ub,nparam,npart,xpsotol,xquicktol,xquicknum,fpsotol,maxpsoit,&
    phipso,newrun,psoseed)

write(*,*) "final parameters = x = ",xpso
write(*,*) "final SMM objective = ",fpso

!This point ends the main program performing the minimization within the structural
!estimation. The following code contains the subfunctions required to perform the task.
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

contains

logical function file_exists(filename)
implicit none

character (len=*) :: filename

inquire(file=filename,exist=file_exists)

end function file_exists

subroutine unctauchen(transarray,grid,znum,zmin,zmax,snum,sigmagrid,rho)
implicit none 

!this routine discretizes the AR(1) productivity processes, subject to 
!unc shocks

!input/output declarations
integer :: znum,snum
double precision :: zmin,zmax,rho
double precision :: transarray(znum,znum,snum),grid(znum),sigmagrid(snum)

!other declarations
integer :: zct,sct,zprimect
double precision :: log_grid(znum),standdev,gridinc

!create grid
call linspace(log_grid,log(zmin),log(zmax),znum)
grid = exp(log_grid)
gridinc = log_grid(2)-log_grid(1)

!loop over unc states
do sct=1,snum
    standdev = sigmagrid(sct)
    do zct=1,znum
        !middle intervals
        do zprimect=2,(znum-1)
            transarray(zct,zprimect,sct) = &
                normcdf(log_grid(zprimect)+gridinc/2.0,rho*log_grid(zct),standdev) - &
                normcdf(log_grid(zprimect)-gridinc/2.0,rho*log_grid(zct),standdev)
        end do !zprimect
        !first interval & last interval take remainder of mass
        transarray(zct,1,sct) = normcdf(log_grid(1)+gridinc/2.0,rho*log_grid(zct),standdev)
        transarray(zct,znum,sct) = 1.0 - normcdf(log_grid(znum)-gridinc/2.0,rho*log_grid(zct),standdev)
    end do !zct
    
    !impose that everything sums to 1 with rounding
    do zct=1,znum
        transarray(zct,:,sct) = transarray(zct,:,sct) / sum(transarray(zct,:,sct))
    end do !zct
    
end do !sct

end subroutine unctauchen

subroutine firmsexogsim(ssimpos,zfirmspos,zfirmsshocks,pr_mat_z,znum,snum,zinit,nfirms,numper)
implicit none

!this subroutine simulates the firm panel of idio prod

!input/output declarations
integer :: znum,snum,zinit,nfirms,numper,ssimpos(numper),zfirmspos(nfirms,numper),zct,zprimect
double precision :: zfirmsshocks(nfirms,numper),pr_mat_z(znum,znum,snum)

!note that zfirmspos is the only output of this subroutine

!other declarations
integer :: t,firmct,ct,sct
double precision :: zsimgrid(znum)

!initialize all of the firms positions to zinit
zfirmspos(:,1) = zinit

!then, simulate each firm for the entire simulation length
do firmct=1,nfirms
do t=2,numper
    
    sct = ssimpos(t); !what is aggregate uncertainty in the last period
    zct = zfirmspos(firmct,t-1); !what was idio prod last period?
    
    zprimect=0; !this will store the new value
    
    !create vector of cumulative probability thresholds from the transition mat
    zsimgrid(:) = 0.0
    zsimgrid(1) = pr_mat_z(zct,1,sct)
    do ct=2,znum
        zsimgrid(ct) = zsimgrid(ct-1) + pr_mat_z(zct,ct,sct)
    end do !ct
    
    !compare U(0,1) draws to cumulative probability thresholds
    if (zfirmsshocks(firmct,t)<zsimgrid(1)) then
        zprimect=1
    else 
        do ct=2,znum
            if (zsimgrid(ct-1)<=zfirmsshocks(firmct,t).and.zfirmsshocks(firmct,t)<zsimgrid(ct)) then
                zprimect=ct
            end if
        end do !ct
    end if
    
    !fill in the simulated position for idio prod
    zfirmspos(firmct,t) = zprimect
        
end do !t
end do !firmct

end subroutine firmsexogsim

subroutine uncexogsim(ssimshocks,asimshocks,ssimpos,asimpos,pr_mat_s,pr_mat_a,ainit,sinit,snum,anum,numper)
implicit none

!this subroutine simulates the uncertainty and aggregate productivity processes

!input/output declarations
integer :: numper,ainit,sinit,snum,anum,ssimpos(numper),asimpos(numper)
double precision :: ssimshocks(numper),asimshocks(numper),pr_mat_s(snum,snum),&
    pr_mat_a(anum,anum,snum)
    
!other declarations
integer :: t,ct,sprimect,aprimect
double precision :: ssimgrid(snum),asimgrid(anum)

!initialize the simulation for uncertainty and for agg prod
ssimpos(1) = sinit
asimpos(1) = ainit

!now simulate the uncertainty process
do t=2,numper
    sprimect=0; !this will store new value
   
    !create vector of cumulative probability thresholds from transition mat
    ssimgrid(1) = pr_mat_s(ssimpos(t-1),1)
    do ct=2,snum
        ssimgrid(ct) = ssimgrid(ct-1) + pr_mat_s(ssimpos(t-1),ct)
    end do !ct

    !compare the uniform shock to the thresholds, and decide which bin
    if (ssimshocks(t)<ssimgrid(1)) then
        sprimect = 1
    else 
        do ct=2,snum
            if (ssimgrid(ct-1)<=ssimshocks(t).and.ssimshocks(t)<ssimgrid(ct)) then
                sprimect = ct
            end if
        end do !ct
    end if

    !fill in the simulated unc position
    ssimpos(t) = sprimect
end do !t

!given the uncertainty process, simulate the aggregate productivity process
do t=2,numper
    aprimect=0; !this will store the new value of agg prod
    
    !create vector of cumulative probability thresholds from transition mat
    !note that the third entry takes uncertainty into account
    asimgrid(1) = pr_mat_a(asimpos(t-1),1,ssimpos(t-1))
    do ct=2,anum
        asimgrid(ct) = asimgrid(ct-1) + pr_mat_a(asimpos(t-1),ct,ssimpos(t-1))
    end do !ct
    
    !compare the uniform shock to the thresholds, and decide which bin
    if (asimshocks(t)<asimgrid(1)) then
        aprimect = 1
    else 
        do ct=2,anum
            if (asimgrid(ct-1)<=asimshocks(t).and.asimshocks(t)<asimgrid(ct)) then
                aprimect = ct
            end if
        end do !ct
    end if
    
    !fill in the simulated unc position
    asimpos(t) = aprimect
end do !t
end subroutine uncexogsim

double precision function y(zval,aval,kval,lval,alpha,nu)
implicit none

!this function evaluates output given parameters and states

double precision :: zval,aval,kval,lval,alpha,nu

y = zval * aval * (kval ** alpha) * (lval ** nu)
    
end function

double precision function ACk(zval,aval,kval,lval,alpha,nu,kprimeval,capirrev,capfix,deltak)
implicit none

!this function evaluates capital AC given parameters and states

!input parameters
double precision :: zval,aval,kval,lval,alpha,nu,kprimeval,capirrev,capfix,deltak

!other parameters
double precision :: yval,changetol,ival

changetol = 1.0e-10
ival = kprimeval - (1.0-deltak) * kval

!start the AC at 0
ACk = 0.0

!take out the partial irreversibility costs
if (ival<-changetol) then
    ACk = ACk - ival * capirrev
end if

!take out the fixed disruption costs
if (abs(ival)>changetol) then
    yval = y(zval,aval,kval,lval,alpha,nu)
    ACk = ACk + yval * capfix
end if

end function ACk

double precision function ACl(zval,aval,kval,lval,alpha,nu,lmin1val,hirelin,firelin,labfix,deltan,wval)
implicit none

!this function evaluates labor AC given parameters and states

!input parameters
double precision :: zval,aval,kval,lval,alpha,nu,lmin1val,hirelin,firelin,labfix,deltan,wval

!other parameters
double precision :: hval,changetol,yval

changetol = 1.0e-10

hval = lval - (1.0-deltan)*lmin1val

!take out the fixed and linear costs if you changed costs
ACl = 0.0
if (abs(hval)>changetol) then
    yval = y(zval,aval,kval,lval,alpha,nu)
    ACl = ACl + labfix * yval; !the fixed costs
    if (hval<-changetol) ACl = ACl - hval * wval * firelin; !the linear firing costs
    if (hval>changetol) ACl = ACl + hval * wval * hirelin; !the linear hiring costs
end if
end function ACl


double precision function RUBC_func(x)
use fixedparam
use macrodata
use microdata
use base_lib
use omp_lib
implicit none

!!!!DECLARATION BLOCK

character(len=*), parameter :: basedir="."

integer :: numstates,numexog,numendog,ct,zct,act,sct,kct,lmin1ct,zprimect,aprimect,sprimect,&
    primect,endogct,exogct,polct,vfct,accelct,polstar,exogprimect,t,numfcst,fcstct,lct,kprimect,ind,piter,GEct,perct,&
    simct,exitflag,smin1ct,firmct,numyrs,yrct,qtrct,&
    numyrdiscard
        
integer, allocatable :: exog_pos(:,:),endog_pos(:,:),loopind(:,:),polmat(:,:,:),polmatold(:,:,:),&
    asimpos(:),ssimpos(:),kbarfcstinds(:,:,:,:),polmatp(:,:),kprime_posp(:,:),lpol_posp(:,:),&
    piterstoresim(:),flagfirms(:,:),endogfirms(:,:),zfirmspos(:,:)

double precision :: start,finish,pval,wval,ajump,zjump,uncpers,uncfreq,sigmaz,sigmaa,zmin,zmax,amin,amax,kmax,&
    lmax,zval,aval,sval,kval,lmin1val,kprimeval,lval,ival,yval,ackval,aclval,vferror,polerror,hval,kbarval,&
    kbarfcstval,pfcstval,wfcstval,Vnextval,weight,perror,Yvalp,Ivalp,ACkvalp,AClvalp,Cvalp,pupdate,Kbarprimevalp,&
    Hvalp,Lvalp,xmean,ymean,x2mean,xymean,xval,kfcsterror,pfcsterror,pvala,pvalb,pvalc,fa,fb,fc,pwidth,KRMSEchange,&
    pRMSEchange,KR2change,pR2change,pmaxDenHaanchange,KmaxDenHaanchange,pavgDenHaanchange,KavgDenHaanchange,TOPkstat,&
    BOTkstat,TOPlstat,BOTlstat,MICROMOM(4),MACROMOM(4),x(6)

double precision, allocatable :: V(:,:,:),asimshocks(:),ssimshocks(:),pr_mat_z(:,:,:),&
    pr_mat_a(:,:,:),pr_mat_s(:,:),pr_mat(:,:),z0(:),k0(:),l0(:),a0(:),sigmazgrid(:),sigmaagrid(:),&
    OUTconstants(:),exog_key(:,:),endog_key(:,:),EVmat(:,:,:),Vold(:,:,:),RHSvec(:),kprime_key(:,:,:),&
    lpol_key(:,:,:),kprime_pos(:,:,:),lpol_pos(:,:,:),distzkl(:,:,:),Ysim(:),Ksim(:),Lsim(:),Isim(:),Hsim(:),&
    ACksim(:),AClsim(:),TOPksim(:),BOTksim(:),TOPlsim(:),BOTlsim(:),ADJksim(:),ADJlsim(:),kbar0(:),&
    kbarfcstmat(:,:,:,:),pfcstmat(:,:,:,:),kfcstcoeff(:,:,:,:),pfcstcoeff(:,:,:,:),wfcstmat(:,:,:,:),&
    Ymat(:,:,:,:),Imat(:,:),ACkmat(:,:,:,:,:),AClmat(:,:,:,:,:),WLmat(:,:,:,:,:),kbarfcstweights(:,:,:,:),&
    psim(:),EVmatp(:,:),Csim(:),Kfcstsim(:),pfcstsim(:),kcoeffstore(:,:,:,:,:),pcoeffstore(:,:,:,:,:),&
    kfcstcoeffnew(:,:,:,:),pfcstcoeffnew(:,:,:,:),kfcsterrorstore(:),pfcsterrorstore(:),pR2store(:,:,:,:),&
    KR2store(:,:,:,:),pRMSEstore(:,:,:,:),KRMSEstore(:,:,:,:),pSSE(:,:,:),KSSE(:,:,:),pSST(:,:,:),KSST(:,:,:),Kmean(:,:,:),&
    pmean(:,:,:),KRMSEchangestore(:),pRMSEchangestore(:),KR2changestore(:),pR2changestore(:),pDenHaanfcst(:),&
    KDenHaanfcst(:),pmaxDenHaanstat(:),KmaxDenHaanstat(:),pavgDenHaanstat(:),KavgDenHaanstat(:),&
    pmaxDenHaanchangestore(:),KmaxDenHaanchangestore(:),pavgDenHaanchangestore(:),KavgDenHaanchangestore(:),&
    zfirmsshocks(:,:),tfpannual(:,:),outputannual(:,:),laborannual(:,:),capitalannual(:,:),TFPsim(:),&
    investmentannual(:,:)

character*5 :: numperstr,numdiscardstr,nfirmsstr,numyrsstr,numyrdiscardstr
character*200 :: matlabstr

!read in data from SMM parameters
ajump = x(1)
zjump = x(2)
uncpers = x(3)
uncfreq = x(4)
sigmaz = x(5)
sigmaa = x(6)

write(*,*) "##########################################"
write(*,*) "SMM OPTIMIZATION"
write(*,*) "PARAMETERS = "
do ct=1,6
write(*,*) x(ct)
end do !ct
!!!!PROGRAM SETUP BLOCK

!call the clock to get the start time and start writing to the log file
start = omp_get_wtime()
open(13,file=basedir//"\RUBC_wrapper.txt")

write(*,*) " "
write(13,*) " "

!!!!END PROGRAM SETUP BLOCK

!!!!!PARAMETER ENTRY BLOCK

!set up the idio prod process
zmin = exp( - 2.0 * ( ( sigmaz ** 2.0 ) / ( 1.0 - rhoz ** 2.0 ) ) ** 0.5 );
zmax = exp( 2.0 * ( ( sigmaz ** 2.0 ) / ( 1.0 - rhoz ** 2.0 ) ) ** 0.5);

!set up the agg prod process
amin = exp( - 2.0 * ( ( sigmaa ** 2.0 ) / ( 1.0 - rhoa ** 2.0 ) ) ** 0.5);
amax = exp( 2.0 * ( ( sigmaa ** 2.0 ) / ( 1.0 - rhoa ** 2.0 ) ) ** 0.5);

!set up the idio capital grid, respecting depreciation
kmax = exp(log(kmin) - dble(knum-1) * log(1.0-deltak))
write(*,"(A,F10.3,A,F10.3,A)") "capital grid = (",kmin,",",kmax,")"
write(13,"(A,F10.3,A,F10.3,A)") "capital grid = (",kmin,",",kmax,")"

!set up the idio labor grid, respecting depreciation
lmax = exp(log(lmin) - dble(lnum-1) * log(1.0-deltan))
write(*,"(A,F10.3,A,F10.3,A)") "labor grid = (",lmin,",",lmax,")"
write(13,"(A,F10.3,A,F10.3,A)") "labor grid = (",lmin,",",lmax,")"

!figure out total number of states
numexog = znum*anum*snum*snum
numendog = knum*lnum
numfcst = kbarnum
numstates = numexog * numendog * numfcst

!control the uncondtional simulation
write(numperstr,'(I5)') numper
write(numdiscardstr,'(I5)') numdiscard

!control the firm panel simulation
write(nfirmsstr,'(I5)') nfirms
numyrs = (numper/4)-1; !note that you always need to have numper (in quarters) divisible by 4
write(numyrsstr,'(I5)') numyrs
numyrdiscard = numdiscard/4; !note that numdiscard must be divisble by 4 as well, i.e. it must be a whole number of years
write(numyrdiscardstr,'(I5)') numyrdiscard

!control the calls to MATLAB to compute GARCH(1,1) and macro moments as well as micro regressions
!and the IQR of the residuals moments
matlabstr = 'matlab -nosplash -nodesktop -r "MACROMOM('&
    //numperstr//','//numdiscardstr//'); '//&
    'MICROMOM('//nfirmsstr//','//numyrsstr//','//numyrdiscardstr//'); quit;"'

!!!!END PARAMETER ENTRY BLOCK

!!!!SOME INITIALIZATIONS/GRID SETUPS

!do allocations for all variable-dimension arrays
allocate(k0(knum),l0(lnum),z0(znum),a0(anum),pr_mat_z(znum,znum,snum),pr_mat_a(anum,anum,snum),pr_mat_s(snum,snum),&
    sigmaagrid(snum),sigmazgrid(snum),OUTconstants(numconstants),exog_key(numexog,4),exog_pos(numexog,4),endog_key(numendog,2),&
    endog_pos(numendog,2),V(numendog,numexog,numfcst),pr_mat(numexog,numexog),EVmat(numexog,numendog,numfcst),&
    loopind(numstates,3),polmat(numendog,numexog,numfcst),Vold(numendog,numexog,numfcst),polmatold(numendog,numexog,numfcst),&
    RHSvec(numendog),kprime_key(numendog,numexog,numfcst),lpol_key(numendog,numexog,numfcst),&
    kprime_pos(numendog,numexog,numfcst),lpol_pos(numendog,numexog,numfcst),asimpos(numper),ssimpos(numper),&
    distzkl(znum,numendog,numper),asimshocks(numper),ssimshocks(numper),Ysim(numper),Ksim(numper),Lsim(numper),Isim(numper),&
    Hsim(numper),ACksim(numper),&
    AClsim(numper),TOPksim(numper),BOTksim(numper),TOPlsim(numper),BOTlsim(numper),ADJksim(numper),ADJlsim(numper),psim(numper),&
    kbar0(kbarnum),&
    kbarfcstmat(anum,snum,snum,numfcst),pfcstmat(anum,snum,snum,numfcst),kfcstcoeff(anum,snum,snum,2),pfcstcoeff(anum,snum,snum,2),&
    wfcstmat(anum,snum,snum,numfcst),Ymat(znum,anum,knum,lnum),Imat(knum,knum),ACkmat(znum,anum,knum,lnum,knum),&
    AClmat(numexog,knum,lnum,lnum,numfcst),WLmat(anum,snum,snum,numfcst,lnum),kbarfcstweights(anum,snum,snum,numfcst),&
    kbarfcstinds(anum,snum,snum,numfcst),EVmatp(znum,numendog),polmatp(znum,numendog),kprime_posp(znum,numendog),&
    lpol_posp(znum,numendog),Csim(numper),Kfcstsim(numper),pfcstsim(numper),kcoeffstore(anum,snum,snum,2,maxGEit),&
    pcoeffstore(anum,snum,snum,2,maxGEit),kfcstcoeffnew(anum,snum,snum,2),pfcstcoeffnew(anum,snum,snum,2),piterstoresim(numper),&
    kfcsterrorstore(maxGEit),pfcsterrorstore(maxGEit),pR2store(anum,snum,snum,maxGEit),KR2store(anum,snum,snum,maxGEit),&
    pRMSEstore(anum,snum,snum,maxGEit),KRMSEstore(anum,snum,snum,maxGEit),&
    pSSE(anum,snum,snum),KSSE(anum,snum,snum),pSST(anum,snum,snum),&
    KSST(anum,snum,snum),Kmean(anum,snum,snum),pmean(anum,snum,snum),KRMSEchangestore(maxGEit),pRMSEchangestore(maxGEit),&
    KR2changestore(maxGEit),pR2changestore(maxGEit),pDenHaanfcst(numper),KDenHaanfcst(numper),pmaxDenHaanstat(maxGEit),&
    KmaxDenHaanstat(maxGEit),pavgDenHaanstat(maxGEit),KavgDenHaanstat(maxGEit),pmaxDenHaanchangestore(maxGEit),&
    KmaxDenHaanchangestore(maxGEit),pavgDenHaanchangestore(maxGEit),KavgDenHaanchangestore(maxGEit),&
    zfirmsshocks(nfirms,numper),&
    tfpannual(nfirms,numyrs),outputannual(nfirms,numyrs),laborannual(nfirms,numyrs),capitalannual(nfirms,numyrs),&
    endogfirms(nfirms,numper),zfirmspos(nfirms,numper),flagfirms(znum,numendog),TFPsim(numper),&
    investmentannual(nfirms,numyrs))

!set up the idiosyncratic capital and labor grids, using the subroutine linspace which replicates MATLAB function
call linspace(k0,log(kmin),log(kmax),knum); k0=exp(k0);
call linspace(l0,log(lmin),log(lmax),lnum); l0=exp(l0);

!set up the aggregate capital grid
call linspace(kbar0,log(kbarmin),log(kbarmax),kbarnum); kbar0 = exp(kbar0);

!set up the sigma or volatility grids and transition matrix
sigmazgrid = (/ sigmaz , zjump*sigmaz /)
sigmaagrid = (/ sigmaa , ajump*sigmaa /)
pr_mat_s(1,:) = (/1.0 - uncfreq , uncfreq /); pr_mat_s(1,:) = pr_mat_s(1,:)/sum(pr_mat_s(1,:))
pr_mat_s(2,:) = (/1.0 - uncpers , uncpers /); pr_mat_s(2,:) = pr_mat_s(2,:)/sum(pr_mat_s(2,:))

!initialize the forecast rules for capital
!initialize the forecast rules for capital
kfcstcoeff(1,1,1,:) = (/   4.6419622828068795E-002 , 0.97142592718050269   /)
kfcstcoeff(1,1,2,:) = (/   8.9961906734298228E-002 , 0.94840064439183314   /)
kfcstcoeff(1,2,1,:) = (/  0.10499059208524510      , 0.93444355249602484    /)
kfcstcoeff(1,2,2,:) = (/   8.3286773019449659E-002 , 0.94753798432733705   /)
kfcstcoeff(2,1,1,:) = (/   9.1711934999120115E-002 , 0.94648818699471937   /)
kfcstcoeff(2,1,2,:) = (/  0.11078131981759207      , 0.93835623468204221   /)
kfcstcoeff(2,2,1,:) = (/  -7.3033510191962440E-003 ,  1.0007832490878241   /)
kfcstcoeff(2,2,2,:) = (/   9.3720909924306969E-002 , 0.94340878611716428   /)
kfcstcoeff(3,1,1,:) = (/   9.0718138203662813E-002 , 0.94788748418664903   /)
kfcstcoeff(3,1,2,:) = (/  0.12620360537051084      , 0.93007463547707414   /)
kfcstcoeff(3,2,1,:) = (/   3.4192432585063064E-003 , 0.99616419551913615   /)
kfcstcoeff(3,2,2,:) = (/  0.12267494292318243      , 0.92791999355260435   /)
kfcstcoeff(4,1,1,:) = (/  0.10552789291615669      , 0.94066484555260443   /)
kfcstcoeff(4,1,2,:) = (/  0.34155291870810395      , 0.80614699576515769   /)
kfcstcoeff(4,2,1,:) = (/   5.6351212545317601E-002 , 0.96607318572086331   /)
kfcstcoeff(4,2,2,:) = (/  0.11964139827109603      , 0.93148312426042579   /)
kfcstcoeff(5,1,1,:) = (/   7.9538515503350879E-002 , 0.95705234881505574   /)
kfcstcoeff(5,1,2,:) = (/   8.4205759999872232E-002 , 0.95709176571491084   /)
kfcstcoeff(5,2,1,:) = (/   6.0812525105999327E-003 , 0.99576182756244358   /)
kfcstcoeff(5,2,2,:) = (/   9.4477205972036807E-002 , 0.94781084855215625   /)

!initialize the forecast rules for price
pfcstcoeff(1,1,1,:) = (/   1.1825963701092423  ,   -0.49454228782951054   /)
pfcstcoeff(1,1,2,:) = (/   1.2615046999472810  ,   -0.52920413604474592   /)
pfcstcoeff(1,2,1,:) = (/   1.0919866133597893  ,   -0.44964836003950076   /)
pfcstcoeff(1,2,2,:) = (/   1.0670044740157572  ,   -0.42642750772903881   /)
pfcstcoeff(2,1,1,:) = (/   1.3061462624135121  ,   -0.57413686521751761   /)
pfcstcoeff(2,1,2,:) = (/   1.3106092296137717  ,   -0.56527210955691376   /)
pfcstcoeff(2,2,1,:) = (/   1.1104300728613385  ,   -0.46608701426028581   /)
pfcstcoeff(2,2,2,:) = (/   1.0983564977097475  ,   -0.45106220617657034   /)
pfcstcoeff(3,1,1,:) = (/   1.2479660920342619  ,   -0.54825748955738640   /)
pfcstcoeff(3,1,2,:) = (/   1.2990712390985699  ,   -0.56637593872754421   /)
pfcstcoeff(3,2,1,:) = (/   1.1210704288290048  ,   -0.47849566325810794   /)
pfcstcoeff(3,2,2,:) = (/   1.0691220951818932  ,   -0.44109849955834290   /)
pfcstcoeff(4,1,1,:) = (/   1.3117202207522143  ,   -0.59239087213133867   /)
pfcstcoeff(4,1,2,:) = (/   1.2289135221955210  ,   -0.53251029547165596   /)
pfcstcoeff(4,2,1,:) = (/   1.0703562190403113  ,   -0.45665156776735755   /)
pfcstcoeff(4,2,2,:) = (/   1.0632175319321056  ,   -0.44561760110697540   /)
pfcstcoeff(5,1,1,:) = (/   1.1358623521273008  ,   -0.49865919675759662   /)
pfcstcoeff(5,1,2,:) = (/   1.2066055316877979  ,   -0.52744522554574425   /)
pfcstcoeff(5,2,1,:) = (/   1.0176031261422835  ,   -0.43241491479560423   /)
pfcstcoeff(5,2,2,:) = (/  0.99331537007904880  ,   -0.41096771371643182   /)
                                                    
!initialize the forecast grids, noting that this can be made more precise later
do act=1,anum
do sct=1,snum
do smin1ct=1,snum
    
    !insert the forecasts into the fcst array
    do fcstct=1,numfcst
        kbarval = kbar0(fcstct); !what is the value of aggregate capital?
        kbarfcstval = exp( kfcstcoeff(act,sct,smin1ct,1) + kfcstcoeff(act,sct,smin1ct,2) * log( kbarval ) ); !implied agg cap fcst
        pfcstval = exp( pfcstcoeff(act,sct,smin1ct,1) + pfcstcoeff(act,sct,smin1ct,2) * log( kbarval ) ); !implied price fcst
        
        !adjust the agg capital forecast to the boundaries of the grid
        kbarfcstval = minval( (/ maxval( (/ kbarmin , kbarfcstval /) ) , kbarmax /) )
        
        !find the interval of kbar0 in which kbarfcst lands, as well as the lin. interp. weight
        ind = fcstct; call hunt(kbar0,numfcst,kbarfcstval,ind)
        weight = ( kbarfcstval - kbar0(ind) ) / ( kbar0(ind+1) - kbar0(ind) )
        
        !got to make sure that out of bounds entries are handled correctly
        if (ind>=1.and.ind<kbarnum) then
            kbarfcstinds(act,sct,smin1ct,fcstct) = ind; kbarfcstweights(act,sct,smin1ct,fcstct) = weight
        else if (ind==0) then
            kbarfcstinds(act,sct,smin1ct,fcstct) = 1; kbarfcstweights(act,sct,smin1ct,fcstct) = 0.0
        else if (ind==kbarnum) then
            kbarfcstinds(act,sct,smin1ct,fcstct) = kbarnum-1; kbarfcstweights(act,sct,smin1ct,fcstct) = 1.0
        end if
        
        !insertion into arrays, including insertion of the implied wage
        kbarfcstmat(act,sct,smin1ct,fcstct) = kbarfcstval
        pfcstmat(act,sct,smin1ct,fcstct) = pfcstval
        wfcstmat(act,sct,smin1ct,fcstct) = theta / pfcstval
        
    end do !fcstct
end do !smin1ct
end do !sct
end do !act

!compute the idio & aggr transition matrices, via Tauchen (1986) adapted to stochastical vol case
call unctauchen(pr_mat_z,z0,znum,zmin,zmax,snum,sigmazgrid,rhoz)
call unctauchen(pr_mat_a,a0,anum,amin,amax,snum,sigmaagrid,rhoa)
 
!index the exogenous variables (z,a,s) & create the unified transition matrix over exogenous variables
ct=0
do zct=1,znum; do act=1,anum; do sct=1,snum; do smin1ct=1,snum;
    ct=ct+1;
    
    !indexing the variables
    exog_pos(ct,1) = zct; exog_key(ct,1) = z0(zct); !insert z
    exog_pos(ct,2) = act; exog_key(ct,2) = a0(act); !insert a
    exog_pos(ct,3) = sct; exog_key(ct,3) = dble(sct); !insert s
    exog_pos(ct,4) = smin1ct; exog_key(ct,4) = dble(smin1ct); !insert s_{-1}
    
    !creating the transition matrix
    do zprimect=1,znum; do aprimect=1,anum; do sprimect=1,snum
        !note that this only indexes numexog/snum of the next period's values, since sct today is smin1ct tomorrow and is FIXED!
        primect=(zprimect-1)*anum*snum*snum + (aprimect-1)*snum*snum + (sprimect-1)*snum + sct
        pr_mat(ct,primect) = pr_mat_z(zct,zprimect,sct) * pr_mat_a(act,aprimect,sct) * pr_mat_s(sct,sprimect)
    end do; end do; end do;
    
    !correction for roundoff error
    pr_mat(ct,:) = pr_mat(ct,:) / sum( pr_mat(ct,:) )
    
end do; end do; end do; end do;

!read in the random shocks
open(8,file="asimshocks.txt")
do ct=1,numper
read(8,*) asimshocks(ct)
end do 
close(8)

open(8,file="ssimshocks.txt")
do ct=1,numper
read(8,*) ssimshocks(ct)
end do 
close(8)

open(8,file="zfirmsshocks.txt")
do firmct=1,nfirms
do t=1,numper
read(8,*) zfirmsshocks(firmct,t)
end do 
end do 
close(8)

!based on the random shocks drawn, simulate the aggregate exogenous processes, occurs outside of fcst loop
call uncexogsim(ssimshocks,asimshocks,ssimpos,asimpos,pr_mat_s,pr_mat_a,ainit,sinit,snum,anum,numper)
call firmsexogsim(ssimpos,zfirmspos,zfirmsshocks,pr_mat_z,znum,snum,zinit,nfirms,numper); !actually perform the simulation

!index the endog variables (k,l_{-1}), noting that the index is used for states and policies
ct=0
do kct=1,knum; do lmin1ct=1,lnum
    ct=ct+1
    endog_pos(ct,1) = kct; endog_key(ct,1) = k0(kct); !insert k
    endog_pos(ct,2) = lmin1ct; endog_key(ct,2) = l0(lmin1ct); !insert l_{-1}
end do; end do;

!indexing for parallelization over V: parallel loops will go over "ct," with
!each thread pulling out endogct, exogct, and fcstct
ct=0
do endogct=1,numendog
do exogct=1,numexog
do fcstct=1,numfcst
    ct=ct+1
    loopind(ct,1) = endogct; loopind(ct,2) = exogct; loopind(ct,3) = fcstct
end do !numfcst
end do !exogct
end do !endogct

!determine and write constants
OUTconstants = (/ dble(znum),dble(knum),dble(lnum),dble(anum),dble(snum),&
    dble(kbarnum),dble(numper),dble(numdiscard),dble(numendog),dble(numexog),dble(numfcst),&
    dble(numstates),deltak,deltan/)


!!!!END INITIALIZATIONS/GRID SETUPS

do GEct=1,maxGEit; !GEct - THIS IS THE INDEX OVER THE NUMBER OF FCST RULE ITERATIONS

!!!!!VF ITERATION BLOCK

write(*,*) " "
write(13,*) " "


!now, create all of the matrices useful later for setting up RHS returns

!output Ymat(z,a,k,l)
!$omp parallel private(zct,act,kct,lct,zval,aval,kval,lval)
!$omp do collapse(4)
do zct=1,znum
do act=1,anum
do kct=1,knum
do lct=1,lnum
    zval = z0(zct); aval=a0(act); kval = k0(kct); lval=l0(lct)
    Ymat(zct,act,kct,lct) = y(zval,aval,kval,lval,alpha,nu)
end do; !lct
end do; !kct
end do; !act
end do; !zct
!$omp end do nowait
!$omp end parallel

!investment Imat(k,k')
do kct=1,knum
do kprimect=1,knum
    kval = k0(kct); kprimeval = k0(kprimect)
    Imat(kct,kprimect) = kprimeval - (1.0-deltak) * kval
end do !kprimect
end do !kct
    
!capital AC ACkmat(z,a,k,l,k')
!$omp parallel private(zct,act,kct,lct,kprimect,zval,aval,kval,lval,kprimeval)
!$omp do collapse(5)
do zct=1,znum
do act=1,anum
do kct=1,knum
do lct=1,lnum
do kprimect=1,knum
    zval = z0(zct); aval=a0(act); kval = k0(kct); lval = l0(lct); kprimeval = k0(kprimect)
    ACkmat(zct,act,kct,lct,kprimect) = &
        ACk(zval,aval,kval,lval,alpha,nu,kprimeval,capirrev,capfix,deltak)
end do !kprimect
end do !lct
end do !kct
end do !act
end do !zct
!$omp end do nowait
!$omp end parallel

!labor AC AClmat(exogct,k,l,l_{-1},K)
!$omp parallel private(zct,act,kct,lct,lmin1ct,sct,fcstct,zval,aval,kval,lval,&
!$omp& lmin1val,wfcstval,smin1ct,exogct)
!$omp do collapse(5)
do exogct=1,numexog
do kct=1,knum
do lct=1,lnum
do lmin1ct=1,lnum
do fcstct=1,numfcst
    zct = exog_pos(exogct,1);
    act = exog_pos(exogct,2);
    sct = exog_pos(exogct,3);
    smin1ct = exog_pos(exogct,4);
    
    zval = z0(zct); aval=a0(act); kval = k0(kct); lval = l0(lct); lmin1val = l0(lmin1ct);
    wfcstval = wfcstmat(act,sct,smin1ct,fcstct);
    AClmat(exogct,kct,lct,lmin1ct,fcstct) = &
        ACl(zval,aval,kval,lval,alpha,nu,lmin1val,hirelin,firelin,labfix,deltan,wfcstval)
        
end do !fcstct
end do !lmin1ct
end do !lct
end do !kct
end do !exogct
!$omp end do nowait
!$omp end parallel

!wage bill WLmat(a,s,s_{-1},K,l)
!$omp parallel private(act,sct,smin1ct,fcstct,lct)
!$omp do collapse(5)
do act=1,anum
do sct=1,snum
do smin1ct=1,snum
do fcstct=1,numfcst
do lct=1,lnum
    WLmat(act,sct,smin1ct,fcstct,lct) = wfcstmat(act,sct,smin1ct,fcstct) * l0(lct)
end do !lct
end do !fcstct
end do !smin1ct
end do !sct
end do !act
!$omp end do
!$omp end parallel

write(*,"(A,F15.1,A)") "Finished setup of return matrices at ",omp_get_wtime()-start," seconds"
write(13,"(A,F15.1,A)") "Finished setup of return matrices at ",omp_get_wtime()-start," seconds"

write(*,*) "Doing VFI."
write(13,*) "Doing VFI."

!initialize the VF and policies
if (GEct==1) then 
    !in this case, initialize using a stupid guess
    Vold(:,:,:) = 0.0;
    V(:,:,:) = 0.0
    polmatold(:,:,:) = numendog/2
    polmat(:,:,:) = 0
    EVmat(:,:,:) = 0.0
else if (GEct>1) then
    !in this case, initialize using last GE iteration
    Vold = V
    polmatold = polmat
    polmat(:,:,:) = 0
    EVmat(:,:,:) = 0.0
end if

!this loop is the "VFI loop" itself, although technically this is policy iteration
do vfct=1,vfmaxit
    
    !!!Howard acceleration step
    do accelct=1,accelmaxit
        
        !set up the parallelization for the Howard acceleration construction
        !$omp parallel private(ct,endogct,exogct,fcstct,zct,act,sct,kct,lmin1ct,polstar,&
        !$omp& kprimect,lct,ind,weight,Vnextval,smin1ct)
        !$omp do
        do ct=1,numstates
            
            !extract states from loop index matrix
            endogct=loopind(ct,1); exogct=loopind(ct,2); fcstct=loopind(ct,3);
            
            !extract positions associated with these states
            zct = exog_pos(exogct,1); act = exog_pos(exogct,2); sct = exog_pos(exogct,3);
            smin1ct=exog_pos(exogct,4)
            kct = endog_pos(endogct,1); lmin1ct= endog_pos(endogct,2); 
            
            !extract policy from polmatold
            polstar = polmatold(endogct,exogct,fcstct)
            
            !extract positions associated with the policies
            kprimect = endog_pos(polstar,1); lct = endog_pos(polstar,2)

            !initialize the RHS by forming the current period return
            V(endogct,exogct,fcstct) = pfcstmat(act,sct,smin1ct,fcstct) * ( Ymat(zct,act,kct,lct) &
                - ACkmat(zct,act,kct,lct,kprimect) - AClmat(exogct,kct,lct,lmin1ct,fcstct) &
                - Imat(kct,kprimect) - WLmat(act,sct,smin1ct,fcstct,lct) )
            
            !find interpolation interval and weights for fcst agg capital
            ind = kbarfcstinds(act,sct,smin1ct,fcstct)
            weight = kbarfcstweights(act,sct,smin1ct,fcstct)
            
            !add discounted expected continuation value
            do exogprimect=1,numexog
              
                !what is the linearly interpolated continuation value?
                Vnextval = weight * Vold(polstar,exogprimect,ind+1) + &
                    (1.0 - weight)*Vold(polstar,exogprimect,ind)
              
                !now, add the discounted expected value associated with this realization
                V(endogct,exogct,fcstct) = V(endogct,exogct,fcstct) + &
                    beta * pr_mat(exogct,exogprimect) * Vnextval
                
            end do !exogprimect
       
        end do !ct
        !$omp end do nowait
        !$omp end parallel
        
        !initialize for the next round of the Howard VFI
        Vold = V

    end do !accelct
    
    
    !!!create EVmat for RHS optimization
    
    !set up the parallelization for the creation of the continuation value
    !$omp parallel private(ct,polct,exogct,fcstct,act,sct,ind,weight,exogprimect,&
    !$omp& Vnextval,smin1ct)
    !$omp do
    do ct=1,numstates
        
        !extract exogenous and policy states
        polct=loopind(ct,1); exogct=loopind(ct,2); fcstct = loopind(ct,3)
        
        !what are act and sct?
        act = exog_pos(exogct,2); sct = exog_pos(exogct,3); smin1ct = exog_pos(exogct,4)
        
        !find interpolation interval and weights for fcst agg capital
        ind = kbarfcstinds(act,sct,smin1ct,fcstct)
        weight = kbarfcstweights(act,sct,smin1ct,fcstct)
              
        !create expected value for the next period based on policy, exogenous transitions, and forecast kbar
        EVmat(exogct,polct,fcstct) = 0.0
        do exogprimect=1,numexog; !loop over exogenous realizations next period
            
            !what is the linearly interpolated continuation value?
            Vnextval = weight * Vold(polct,exogprimect,ind+1) + &
                    (1.0 - weight)*Vold(polct,exogprimect,ind)

            !add the weighted continuation value to the next period
            EVmat(exogct,polct,fcstct) = EVmat(exogct,polct,fcstct) + pr_mat(exogct,exogprimect) * Vnextval
        end do !exogprimect
        
    end do !ct
    !$omp end do nowait
    !$omp end parallel

    !!!RHS optimization step
    
    !set up the parallelization for the optimization of the Bellman RHS
    !$omp parallel private(ct,endogct,exogct,fcstct,zct,act,sct,kct,lmin1ct,&
    !$omp& polct,kprimect,lct,RHSvec,polstar,smin1ct) 
    !$omp do
    do ct=1,numstates
        
       !extract states from loop index matrix
        endogct=loopind(ct,1); exogct=loopind(ct,2); fcstct=loopind(ct,3);
        
        !extract positions associated with these states
        zct = exog_pos(exogct,1); act = exog_pos(exogct,2); sct = exog_pos(exogct,3);
        smin1ct=exog_pos(exogct,4)
        kct = endog_pos(endogct,1); lmin1ct= endog_pos(endogct,2); 
        
        !create RHS vector by looping over policies
        do polct=1,numendog
            
            !extract positions associated with the policies
            kprimect = endog_pos(polct,1); lct = endog_pos(polct,2)

            !initialize the RHS by forming the current period return
            RHSvec(polct) = pfcstmat(act,sct,smin1ct,fcstct) * ( Ymat(zct,act,kct,lct) &
                - ACkmat(zct,act,kct,lct,kprimect) - AClmat(exogct,kct,lct,lmin1ct,fcstct) &
                - Imat(kct,kprimect) - WLmat(act,sct,smin1ct,fcstct,lct) )
              
            !actually form the RHS by adding continuation value
            RHSvec(polct) = RHSvec(polct) + beta * EVmat(exogct,polct,fcstct)
            
        end do !polct
        
        !extract policies via selection of max payoff
        polstar = maxloc(RHSvec,1)
        polmat(endogct,exogct,fcstct) = polstar
        V(endogct,exogct,fcstct) = RHSvec(polstar)
        
    end do !ct
    !$omp end do nowait
    !$omp end parallel
    
    !compute errors and output the info
    vferror = maxval(abs(V-Vold))
    polerror = maxval(abs(polmat-polmatold))
    
    write(13,"(A,I3,A,F15.5,A)") "VF iter = ",vfct," in ",omp_get_wtime()-start," seconds"
    write(13,"(A,I3,A,F15.5)") "VF iter = ",vfct,", VF error = ",vferror
    write(13,"(A,I3,A,F15.5)") "VF iter = ",vfct,", policy error = ",polerror
    write(13,*) " "
    
    write(*,"(A,I3,A,F15.5,A)") "VF iter = ",vfct," in ",omp_get_wtime()-start," seconds"
    write(*,"(A,I3,A,F15.5)") "VF iter = ",vfct,", VF error = ",vferror
    write(*,"(A,I3,A,F15.5)") "VF iter = ",vfct,", policy error = ",polerror
    write(*,*) " "
    
    !exit criterion, then initialize for the next round of VFI
    if (polerror<vferrortol) exit
    Vold = V
    polmatold = polmat
    
end do !vfmaxit

!convert polmat to policies for capital and labor
do endogct=1,numendog
do exogct=1,numexog
do fcstct=1,numfcst
    !convert policies to actual capital and labor values
    kprime_key(endogct,exogct,fcstct) = endog_key(polmat(endogct,exogct,fcstct),1)
    kprime_pos(endogct,exogct,fcstct) = endog_pos(polmat(endogct,exogct,fcstct),1)
    
    !convert policies to capital and labor integer grid points
    lpol_key(endogct,exogct,fcstct) = endog_key(polmat(endogct,exogct,fcstct),2)
    lpol_pos(endogct,exogct,fcstct) = endog_pos(polmat(endogct,exogct,fcstct),2)
end do !fcstct
end do !exogct
end do !endogct

write(*,*) " "
write(13,*) " "

!!!!!END VF ITERATION BLOCK

!!!!!SIMULATION BLOCK

write(*,*) "Doing unconditional simulation."
write(13,*) "Doing unconditional simulation."

!first, initialize the distribution over endogenous variables
distzkl(:,:,:) = 0.0; !set everything to zero

!chosen point to center the mass
kct = 50
lmin1ct=20
endogct = (kct-1)*lnum + lmin1ct

distzkl(:,(endogct-5):(endogct+5),1) = 1.0

!now, round the dist
distzkl(:,:,1) = distzkl(:,:,1) / sum(distzkl(:,:,1))

!then, initialize the aggregate series to 0
Ysim(:) = 0.0; Ksim = Ysim; Lsim = Ysim; Isim = Ysim; Hsim = Ysim; ACksim = Ysim; AClsim = Ysim;
TOPksim = Ysim; BOTksim = Ysim; TOPlsim = Ysim; BOTlsim = Ysim; ADJksim = Ysim; ADJlsim = Ysim;
Csim = Ysim; psim =Ysim;

!start with aggregate capital guessed at some reasonable value
Ksim(1) = (kbarmin + kbarmax)/dble(2.0)

!initialize firm panel arrays
flagfirms(:,:) = 0; !everything to 0
endogfirms(:,:) = 0; !everything to 0

endogfirms(:,1) = numendog/2; !initialize to the middle of the endogenous variables grid for all firms
flagfirms(zinit,numendog/2) = 1; !you need to calculate policies at points in flagfirms

write(*,"(A,F15.1,A)") "Starting actual sim at ",omp_get_wtime()-start," seconds."    
write(13,"(A,F15.1,A)") "Starting actual sim at ",omp_get_wtime()-start," seconds."

!now, loop over periods in the simulation, tracking the endogenous
!movement of weight around the grid
do t=1,numper-1
    
    !aggregate states
    if (t>1) then
        !in this case you don't get a segmentation fault when looking for s_{-1} index
        act = asimpos(t); sct=ssimpos(t); smin1ct = ssimpos(t-1)
    else if (t==1) then
        !initialize the very first one to low unc previously
        act = asimpos(t); sct=ssimpos(t); smin1ct = 1;
    end if
    aval = a0(act)
    
    !what is the next period forecast capital, in grid boundaries? index and weight?
    kbarfcstval = exp( kfcstcoeff(act,sct,smin1ct,1) + kfcstcoeff(act,sct,smin1ct,2) * log( Ksim(t) ) );
    kbarfcstval = minval( (/ maxval( (/ kbarmin , kbarfcstval /) ) , kbarmax /) )   
    ind = kbarnum/2; call hunt(kbar0,numfcst,kbarfcstval,ind)
    weight = ( kbarfcstval - kbar0(ind) ) / ( kbar0(ind+1) - kbar0(ind) )
    
    !guard against off grid point values
    if (ind==kbarnum) then 
        ind = kbarnum-1;
        weight = 1.0;
    else if (ind==0) then
        ind = 1;
        weight = 0.0;
    end if
    
    Kfcstsim(t+1) = kbarfcstval
    
    !what is the current period forecast price?
    pfcstsim(t) = exp( pfcstcoeff(act,sct,smin1ct,1) + pfcstcoeff(act,sct,smin1ct,2) * log( Ksim(t) ) );
    
    
    !note that EVmat doesn't depend on p, so it can be pre-computed now, as EVmatp
    
    !set up the parallelization for the creation of the continuation value
    !$omp parallel private(zct,polct,exogct,exogprimect,Vnextval)
    !$omp do collapse(2)
    do zct=1,znum
    do polct=1,numendog
        
        !what is the implied value of exogct, given simulated aggregate states?
        exogct = (zct-1)*anum*snum*snum + (act-1)*snum*snum + (sct-1)*snum  + smin1ct
              
        !create expected value for the next period based on policy, exogenous transitions,
        !and forecast capital
        EVmatp(zct,polct) = 0.0
        do exogprimect=1,numexog; !loop over exogenous realizations next period
            
            !what is the linearly interpolated continuation value?
            Vnextval = weight * V(polct,exogprimect,ind+1) + &
                    (1.0 - weight)*V(polct,exogprimect,ind)

            !add the weighted continuation value to the next period
            EVmatp(zct,polct) = EVmatp(zct,polct) + pr_mat(exogct,exogprimect) * Vnextval
            
        end do !exogprimect
    
    end do !zct
    end do !endogct
    !$omp end do nowait
    !$omp end parallel

    !set up the boundaries of the bisection
    if (t>=25) then
        pvala = pfcstsim(t)-pwindow !lower boundary
        pvalb =  pfcstsim(t) + pwindow !upper boundary
        pvalc = pvala + dble(0.67) * (pvalb-pvala) 
    else if (t<25) then
        pvala = plb
        pvalb = pub
        pvalc = pvala + dble(0.67) * (pvalb-pvala) 
    end if

    !iterate over the market-clearing value of p, for each value
    !redoing the optimization of the RHS of the Bellman equation
    !and recomputing policies
    do piter=1,maxpit
        !initialize the price-dependent policies to 0
        polmatp(:,:) = 0; kprime_posp(:,:) = 0; lpol_posp(:,:) = 0
        
        !brent setup
        if (piter==1) pval = pvala
        if (piter==2) pval = pvalb
        if (piter==3) pval = pvalc
        
        !brent restart
        if (piter==pcutoff) then 
            pvala = pfcstsim(t)-4.0*pwindow;
            pval = pvala;
        else if (piter==pcutoff+1) then 
            pvalb =  pfcstsim(t)+4.0*pwindow;
            pval = pvalb
        else if (piter==pcutoff+2) then
            pvalc = pvala + dble(0.67) * (pvalb-pvala) 
            pval = pvalc
        end if
        
        !if not restarting or initializing
        if ((piter>3.and.piter<pcutoff).or.(piter>pcutoff+2)) then 
            
        
            !first, try inverse quadratic interpolation of the excess demand function
            pval = ( pvala * fb * fc ) / ( (fa - fb) * (fa - fc) ) &
                + ( pvalb * fa * fc ) / ( (fb - fa) * (fb - fc ) ) &
                + ( pvalc * fa * fb ) / ( (fc - fa) * (fc - fb ) )
       
            !if it lies within bounds, and isn't too close to the bounds, then done
            
            !o/w, take bisection step
            if ((minval( (/ abs(pvala - pval), abs(pvalb-pval) /) )<&
                    abs( (pvalb-pvala)/dble(9.0) ) ).or.(pval<pvala).or.(pval>pvalb))   then
                pval = (pvala + pvalb) / dble(2.0)
         
            end if
            
        end if 
        
        
        wval = theta / pval; !what is the wage implied by this value of p?
        
        !this block computes the policies given the current pval & wval at each
        !point (z,kl_{-1})
        
        Yvalp = 0.0; 
        Ivalp = 0.0; 
        ACkvalp = 0.0; 
        AClvalp = 0.0; 
        Kbarprimevalp = 0.0; 
        Hvalp = 0.0;
        Lvalp = 0.0
        
        !$omp parallel private(zct,endogct,exogct,kct,lmin1ct,RHSvec,polct,kprimect,&
        !$omp& lct,polstar) reduction(+:Yvalp,Ivalp,ACkvalp,AClvalp,Kbarprimevalp,Hvalp,&
        !$omp& Lvalp)
        !$omp do collapse(2)
        do zct=1,znum
        do endogct=1,numendog
            !restriction on which policies to compute - important for time
            if (distzkl(zct,endogct,t)>disttol) then                        
            !extract states from loop indexes
            exogct = (zct-1)*anum*snum*snum + (act-1)*snum*snum + (sct-1)*snum  + smin1ct
            
            !extract positions associated with these states
            kct = endog_pos(endogct,1);
            lmin1ct= endog_pos(endogct,2); 
            
            !create RHS vector by looping over policies
            RHSvec(:) = 0.0
            do polct=1,numendog
                
                !extract positions associated with the policies
                kprimect = endog_pos(polct,1); lct = endog_pos(polct,2)

                !initialize the RHS by forming the current period return
                RHSvec(polct) = pval * ( Ymat(zct,act,kct,lct) &
                    - ACkmat(zct,act,kct,lct,kprimect) &
                    - Imat(kct,kprimect) - wval * l0(lct) &
                    - ACl(z0(zct),a0(act),k0(kct),l0(lct),alpha,nu,l0(lmin1ct),hirelin,firelin,&
                    labfix,deltan,wval))
                  
                !actually form the RHS by adding continuation value
                RHSvec(polct) = RHSvec(polct) + beta * EVmatp(zct,polct)
               
            end do !polct
                
            !what is the new optimum value in index form and in capital and labor indexes?
            polmatp(zct,endogct) =  maxloc(RHSvec,1)
            kprime_posp(zct,endogct) = endog_pos(polmatp(zct,endogct),1)
            lpol_posp(zct,endogct) = endog_pos(polmatp(zct,endogct),2)
            
            !what are the investment and labor for this position?
            lct = lpol_posp(zct,endogct); !current labor, policy
            kprimect = kprime_posp(zct,endogct); !next period capital, policy
            
            !what are the aggregates?
            Yvalp = Yvalp + distzkl(zct,endogct,t) * Ymat(zct,act,kct,lct)
            Ivalp = Ivalp + distzkl(zct,endogct,t) * Imat(kct,kprimect)
            ACkvalp = ACkvalp + distzkl(zct,endogct,t) * ACkmat(zct,act,kct,lct,kprimect)
            AClvalp = AClvalp + distzkl(zct,endogct,t) * ACl(z0(zct),a0(act),k0(kct),&
                l0(lct),alpha,nu,l0(lmin1ct),hirelin,firelin,labfix,deltan,wval)
            Kbarprimevalp = Kbarprimevalp + distzkl(zct,endogct,t) * k0(kprimect)
            Hvalp = Hvalp + distzkl(zct,endogct,t) * ( l0(lct) - (1.0-deltan) * l0(lmin1ct))
            Lvalp = Lvalp + distzkl(zct,endogct,t) * l0(lct)
        
            end if !disttol
        
        end do !endogct
        end do !zct
        !$omp end do nowait
        !$omp end parallel
        
        !what is implied consumption?
        Cvalp = Yvalp - Ivalp - ACkvalp - AClvalp
        
        !are you initializing the brent?
        if (piter==1) fa = (1/pval) - Cvalp
        if (piter==2) fb = (1/pval) - Cvalp
        if (piter==3) fc = (1/pval) - Cvalp
        
        !are you restarting the brent?
        if (piter==pcutoff) fa = (1/pval) - Cvalp
        if (piter==pcutoff+1) fb = (1/pval) - Cvalp
        if (piter==pcutoff+2) fc = (1/pval) - Cvalp
        
        !what is the error given by this implied consumption?
        perror = (1/pval) - Cvalp
        
        !if not restarting or initializing
        if ((piter>3.and.piter<pcutoff).or.(piter>pcutoff+2)) then 
            if (perror<0) then
                pvalc = pvalb; fc = fb;
                pvalb = pval; fb = perror;
                !pval a doesn't change
            else if (perror>=0) then
                pvalc = pvala; fc = fa;
                pvala = pval; fa = perror;
                !pval b doesn't change
            end if
        end if
      
        !exit criterion for market-clearing
        if (abs(perror)<perrortol.and.piter>2) exit
      
        end do !piter
       
    !record number of price iterations
    piterstoresim(t) = piter-1
    
    !output price stats on certain periods
    if (mod(t,500)==1) then 
        write(*,"(A,I5,A,I5,A,F10.4,A,F10.4)") "t = ",t,", piter = ",piter-1,&
        ", pval = ",pval,", perror = ",perror
        
        write(13,"(A,I5,A,I5,A,F10.4,A,F10.4)") "t = ",t,", piter = ",piter-1,&
        ", pval = ",pval,", perror = ",perror
        
    end if
    
    !insert market-clearing price and other stuff into sim series
    psim(t) = pval; 
    Csim(t) = Cvalp
    Ysim(t) = Yvalp
    Isim(t) = Ivalp
    ACksim(t) = ACKvalp
    AClsim(t) = AClvalp
    Ksim(t+1) = Kbarprimevalp
    Hsim(t) = Hvalp
    Lsim(t) = Lvalp
    
    !now that the market-clearing price is determined, move on to insert weight into the next period
    do zct=1,znum
    do endogct=1,numendog
        if (distzkl(zct,endogct,t)>disttol) then       
        !based on the latest price, what is the policy here?
        polstar = polmatp(zct,endogct)
        
        !insert distributional weight in appropriate slots next period
        distzkl(:,polstar,t+1) = distzkl(:,polstar,t+1) + pr_mat_z(zct,:,sct)*distzkl(zct,endogct,t)
        end if 
    end do !endogct
    end do !zct
    
    !now, round to make sure that you're ending up with a distribution which makes sense each period
    distzkl(:,:,t+1) = distzkl(:,:,t+1)/sum(distzkl(:,:,t+1))
   
   
   
    !now, do the firm simulation
    
    !need to loop over states and compute policies in any points where you didn't above
    !$omp parallel private(zct,endogct,exogct,kct,lmin1ct,RHSvec,polct,kprimect,lct)
    !$omp do collapse(2)
    do zct=1,znum
    do endogct=1,numendog
        !restriction on which policies to compute that HAVEN't been computed above - important for time
        if ((flagfirms(zct,endogct)>0).and.(distzkl(zct,endogct,t)<=disttol)) then                        
            
            !extract states from loop indexes
            exogct = (zct-1)*anum*snum*snum + (act-1)*snum*snum + (sct-1)*snum  + smin1ct
            
            !extract positions associated with these states
            kct = endog_pos(endogct,1);
            lmin1ct= endog_pos(endogct,2); 
            
            !create RHS vector by looping over policies
            RHSvec(:) = 0.0
            do polct=1,numendog
                
                !extract positions associated with the policies
                kprimect = endog_pos(polct,1); lct = endog_pos(polct,2)

                !initialize the RHS by forming the current period return
                RHSvec(polct) = pval * ( Ymat(zct,act,kct,lct) &
                    - ACkmat(zct,act,kct,lct,kprimect) &
                    - Imat(kct,kprimect) - wval * l0(lct) &
                    - ACl(z0(zct),a0(act),k0(kct),l0(lct),alpha,nu,l0(lmin1ct),hirelin,firelin,&
                    labfix,deltan,wval))
                  
                !actually form the RHS by adding continuation value
                RHSvec(polct) = RHSvec(polct) + beta * EVmatp(zct,polct)
               
            end do !polct
                
            !what is the new optimum value in index form and in capital and labor indexes?
            polmatp(zct,endogct) =  maxloc(RHSvec,1)
            kprime_posp(zct,endogct) = endog_pos(polmatp(zct,endogct),1)
            lpol_posp(zct,endogct) = endog_pos(polmatp(zct,endogct),2)
            
        end if !disttol
    
    end do !endogct
    end do !zct
    !$omp end do nowait
    !$omp end parallel

    !now, actually do the iteration for the firm simulation
    flagfirms(:,:) = 0; !assume policy doesn't need to be computed next period
    do firmct=1,nfirms
        
        !what are idio exog and endog states for this firm?
        zct=zfirmspos(firmct,t)
        zprimect=zfirmspos(firmct,t+1)
        endogct = endogfirms(firmct,t)

        
        !what is the policy for this firm?
        polstar = polmatp(zct,endogct)

        !record policy in the endogenous states array for the next period
        endogfirms(firmct,t+1) = polstar
        
        !does the policy need to be computed for this firm next period?
        flagfirms(zprimect,polstar) = flagfirms(zprimect,polstar) + 1
        
        
    end do !firmct
      
end do !t

write(*,*) " "
write(*,"(A,F15.1,A)") "Finished dist manipulation in ",omp_get_wtime()-start," seconds."    

write(13,*) " "
write(13,"(A,F15.1,A)") "Finished dist manipulation in ",omp_get_wtime()-start," seconds."    

if (checkbounds==1) then
    
    write(*,*) " "
    write(*,*) "Checking for bounds of state space in simulation."
    
    write(13,*) " "
    write(13,*) "Checking for bounds of state space in simulation."
    
    TOPksim(:) = 0.0
    BOTksim(:) = 0.0
    TOPlsim(:) = 0.0 
    BOTlsim(:) = 0.0
    
    do t = 1,numper-1
        do zct=1,znum
        do endogct=1,numendog
            kct = endog_pos(endogct,1); lmin1ct = endog_pos(endogct,2);
            if (kct==1) BOTksim(t) = BOTksim(t) + distzkl(zct,endogct,t)
            if (lmin1ct==1) BOTlsim(t) = BOTlsim(t) + distzkl(zct,endogct,t)
            if (kct==knum) TOPksim(t) = TOPksim(t) + distzkl(zct,endogct,t)
            if (lmin1ct==lnum) TOPlsim(t) = TOPlsim(t) + distzkl(zct,endogct,t)
        end do !endogct
        end do !zct
        
    end do !t


TOPkstat = sum(TOPksim((numdiscard+1):(numper-1)))/(numper-numdiscard-1)
BOTkstat = sum(BOTksim((numdiscard+1):(numper-1)))/(numper-numdiscard-1)
TOPlstat = sum(TOPlsim((numdiscard+1):(numper-1)))/(numper-numdiscard-1)
BOTlstat = sum(BOTlsim((numdiscard+1):(numper-1)))/(numper-numdiscard-1)

end if !checkbounds
write(*,*) " "
write(*,"(A,F15.1,A)") "Finished simulation in ",omp_get_wtime()-start," seconds."    

write(13,*) " "
write(13,"(A,F15.1,A)") "Finished simulation in ",omp_get_wtime()-start," seconds."

!!!!!END SIMULATION BLOCK

!!!!START THE FCST RULE UPDATE BLOCK

!first, keep a record of the fcst rule coefficients for capital and price
kcoeffstore(:,:,:,:,GEct) = kfcstcoeff(:,:,:,:)
pcoeffstore(:,:,:,:,GEct) = pfcstcoeff(:,:,:,:)

!do the OLS estimation to update the fcst rules

do act=1,anum
do sct=1,snum
do smin1ct=1,snum
    
    !do kprime first
    xmean = 0.0
    ymean = 0.0
    x2mean = 0.0
    xymean = 0.0
    perct=0
    do t = (numdiscard+1),(numper-1)
        if ((asimpos(t)==act.and.ssimpos(t)==sct).and.ssimpos(t-1)==smin1ct) then
            perct = perct+1
            xval = log(Ksim(t))
            yval = log(Ksim(t+1))
            
            xmean = xmean + xval
            ymean = ymean + yval
            x2mean = x2mean + (xval**2.0)
            xymean = xymean + (xval * yval)
            
        end if
    end do !t
    
    xmean = xmean / dble(perct)
    ymean = ymean / dble(perct)
    x2mean = x2mean / dble(perct)
    xymean = xymean / dble(perct)
        
    !insert the implied OLS coefficients into the coefficient storage
    kfcstcoeffnew(act,sct,smin1ct,2) = ( xymean - xmean * ymean ) / ( x2mean - ( xmean ** 2.0 ) )
    kfcstcoeffnew(act,sct,smin1ct,1) = ymean - kfcstcoeffnew(act,sct,smin1ct,2) * xmean
    
    !then do price
    xmean = 0.0
    ymean = 0.0
    x2mean = 0.0
    xymean = 0.0
    perct=0
    do t = (numdiscard+1),(numper-1)
        if ((asimpos(t)==act.and.ssimpos(t)==sct).and.ssimpos(t-1)==smin1ct) then
            perct = perct+1
            xval = log(Ksim(t))
            yval = log(psim(t))
            
            xmean = xmean + xval
            ymean = ymean + yval
            x2mean = x2mean + (xval**2.0)
            xymean = xymean + (xval * yval)
            
        end if
    end do !t
    
    xmean = xmean / dble(perct)
    ymean = ymean / dble(perct)
    x2mean = x2mean / dble(perct)
    xymean = xymean / dble(perct)
    
    !insert the implied OLS coefficients into the coefficient storage
    pfcstcoeffnew(act,sct,smin1ct,2) = ( xymean - xmean * ymean ) / ( x2mean - ( xmean ** 2.0 ) )
    pfcstcoeffnew(act,sct,smin1ct,1) = ymean - pfcstcoeffnew(act,sct,smin1ct,2) * xmean
    
end do !smin1ct
end do !sct
end do !act

!compute the error - how far apart are assumed and eqbm OLS estimates?
kfcsterror = maxval(abs(kfcstcoeff-kfcstcoeffnew))
pfcsterror = maxval(abs(pfcstcoeff-pfcstcoeffnew))

!insert errors into storage
kfcsterrorstore(GEct) = kfcsterror
pfcsterrorstore(GEct) = pfcsterror

!!!!compute RMSE and Rsquared for the forecast rules
pSSE(:,:,:) = 0.0
KSSE(:,:,:) = 0.0
pSST(:,:,:) = 0.0
KSST(:,:,:) = 0.0
Kmean(:,:,:) = 0.0
pmean(:,:,:) = 0.0

!loop over aggregate states
do act=1,anum
do sct=1,snum
do smin1ct=1,snum
    !find mean and fcst errors first for price
    perct = 0
    do t=(numdiscard+1),(numper-1)
        if ((asimpos(t)==act.and.ssimpos(t)==sct).and.ssimpos(t-1)==smin1ct) then
            perct = perct+1
            pmean(act,sct,smin1ct) = pmean(act,sct,smin1ct) + log(psim(t))
            pSSE(act,sct,smin1ct) = pSSE(act,sct,smin1ct) + ( log(psim(t)) - log(pfcstsim(t)) )**dble(2.0)
        end if !act and sct flag
    end do !t
    
    !normalize mean, compute RMSE for price
    pmean(act,sct,smin1ct) = pmean(act,sct,smin1ct)/dble(perct)
    pRMSEstore(act,sct,smin1ct,GEct) = sqrt( pSSE(act,sct,smin1ct) / dble(perct) )
    
    !then for capital
    perct = 0
    do t=(numdiscard+1),(numper-1)
        if ((asimpos(t)==act.and.ssimpos(t)==sct).and.ssimpos(t-1)==smin1ct) then
            perct = perct+1
            Kmean(act,sct,smin1ct) = Kmean(act,sct,smin1ct) + log(Ksim(t+1))
            KSSE(act,sct,smin1ct) = KSSE(act,sct,smin1ct) + ( log(Ksim(t+1)) - log(Kfcstsim(t+1)) )**dble(2.0)
        end if !act and sct flag
    end do !t
    
    !normalize mean, compute RMSE for capital
    Kmean(act,sct,smin1ct) = Kmean(act,sct,smin1ct)/dble(perct)
    KRMSEstore(act,sct,smin1ct,GEct) = sqrt( KSSE(act,sct,smin1ct) / dble(perct) )
    
    !now compute the total sum squares
    do t=(numdiscard+1),(numper-1)
        if ((asimpos(t)==act.and.ssimpos(t)==sct).and.ssimpos(t-1)==smin1ct) then
            pSST(act,sct,smin1ct) = pSST(act,sct,smin1ct) + (log(psim(t)) - pmean(act,sct,smin1ct))**dble(2.0)
            KSST(act,sct,smin1ct) = KSST(act,sct,smin1ct) + (log(Ksim(t+1)) - Kmean(act,sct,smin1ct))**dble(2.0)
        end if !act and sct flag
    end do !t
    
    !now, compute the R^2 values
    pR2store(act,sct,smin1ct,GEct) = 1.0 - ( pSSE(act,sct,smin1ct) / pSST(act,sct,smin1ct) )
    KR2store(act,sct,smin1ct,GEct) = 1.0 - ( KSSE(act,sct,smin1ct) / KSST(act,sct,smin1ct) )
end do !smin1ct
end do !sct
end do !act



!!!record the change of the RMSE's and the R^2's
if (GEct==1) then
    
    !in this case, it isn't change, just absolute RMSE and distance of R^2 from 1
    KRMSEchange = maxval(abs(KRMSEstore(:,:,:,GEct)))
    pRMSEchange = maxval(abs(pRMSEstore(:,:,:,GEct)))
    KR2change = maxval(abs(1 - KR2store(:,:,:,GEct)))
    pR2change = maxval(abs(1 - pR2store(:,:,:,GEct)))

else 

    !now, it's the change in the metrics
    KRMSEchange = maxval(abs(KRMSEstore(:,:,:,GEct)-KRMSEstore(:,:,:,GEct-1)))
    pRMSEchange = maxval(abs(pRMSEstore(:,:,:,GEct)-pRMSEstore(:,:,:,GEct-1)))
    KR2change = maxval(abs(KR2store(:,:,:,GEct)-KR2store(:,:,:,GEct-1)))
    pR2change = maxval(abs(pR2store(:,:,:,GEct)-pR2store(:,:,:,GEct-1)))

end if

!store the RMSE and R2 errors
KRMSEchangestore(GEct) = KRMSEchange
pRMSEchangestore(GEct) = pRMSEchange
KR2changestore(GEct) = KR2change
pR2changestore(GEct) = pR2change


!!!Compute the Den Haan fcst series

!initialize the entire fcst series
pDenHaanfcst(:) = 0.0
KDenHaanfcst(:) = 0.0

!initialize the period before the sample
pDenHaanfcst(numdiscard) = log(psim(numdiscard))
KDenHaanfcst(numdiscard) = log(Ksim(numdiscard))

!initialize the statistics
pmaxDenHaanstat(GEct) = 0.0
KmaxDenHaanstat(GEct) = 0.0
pavgDenHaanstat(GEct) = 0.0
KavgDenHaanstat(GEct) = 0.0

!loop over the periods
do t=(numdiscard+1),(numper-1)
    
    !get states from last period, when the capital forecast was being made
    act = asimpos(t-1)
    sct = ssimpos(t-1)
    smin1ct = ssimpos(t-2)
    
    !insert the capital fcst
    KDenHaanfcst(t) = kfcstcoeff(act,sct,smin1ct,1) + kfcstcoeff(act,sct,smin1ct,2)*KDenHaanfcst(t-1)
    
    !now, get states from this period, when the price forecast is being made
    act = asimpos(t)
    sct = ssimpos(t)
    smin1ct = ssimpos(t-1)
    
    !insert the price fcst
    pDenHaanfcst(t) = pfcstcoeff(act,sct,smin1ct,1) + pfcstcoeff(act,sct,smin1ct,2)*KDenHaanfcst(t)
    
    !iterate the max statistics
    pmaxDenHaanstat(GEct) = maxval((/pmaxDenHaanstat(GEct), abs(pDenHaanfcst(t) - log(psim(t))) /))
    KmaxDenHaanstat(GEct) = maxval((/KmaxDenHaanstat(GEct), abs(KDenHaanfcst(t) - log(Ksim(t))) /))
    
    !iterate the average statistics
    pavgDenHaanstat(GEct) = pavgDenHaanstat(GEct) + abs(pDenHaanfcst(t) - log(psim(t)))
    KavgDenHaanstat(GEct) = KavgDenHaanstat(GEct) + abs(KDenHaanfcst(t) - log(Ksim(t)))
    
end do !t

!normalize the average statistics
pavgDenHaanstat(GEct) = pavgDenHaanstat(GEct) / (dble(numper-numdiscard-1))
KavgDenHaanstat(GEct) = KavgDenHaanstat(GEct) / (dble(numper-numdiscard-1))


!record the change of the avg and max statistics
if (GEct==1) then

    pmaxDenHaanchange = pmaxDenHaanstat(GEct)
    KmaxDenHaanchange = KmaxDenHaanstat(GEct)
    pavgDenHaanchange = pavgDenHaanstat(GEct)
    KavgDenHaanchange = KavgDenHaanstat(GEct)

else if (GEct>1) then

    pmaxDenHaanchange = abs(pmaxDenHaanstat(GEct)-pmaxDenHaanstat(GEct-1))
    KmaxDenHaanchange = abs(KmaxDenHaanstat(GEct)-KmaxDenHaanstat(GEct-1))
    pavgDenHaanchange = abs(pavgDenHaanstat(GEct)-pavgDenHaanstat(GEct-1))
    KavgDenHaanchange = abs(KavgDenHaanstat(GEct)-KavgDenHaanstat(GEct-1))

end if
    
!store the changes    
pmaxDenHaanchangestore(GEct) = pmaxDenHaanchange
KmaxDenHaanchangestore(GEct) = KmaxDenHaanchange
pavgDenHaanchangestore(GEct) = pavgDenHaanchange
KavgDenHaanchangestore(GEct) = KavgDenHaanchange

!display the old and the new coefficients
write(*,*) " "
write(*,"(A,I5)") "GE Iteration ",GEct
write(*,"(A,F10.5,F10.5)") "Fcst Coeff Error (capital, price) = ",kfcsterror,pfcsterror
write(*,"(A,F10.5,F10.5)") "Fcst RMSE Change (capital, price) = ",KRMSEchange,pRMSEchange
write(*,"(A,F10.5,F10.5)") "Fcst R2 Change = ",KR2change,pR2change
write(*,"(A,F10.5,F10.5)") "Capital Den Haan % Error (avg,max) = ",KavgDenHaanstat(GEct),KmaxDenHaanstat(GEct)
write(*,"(A,F10.5,F10.5)") "Price Den Haan % Error (avg,max) = ",pavgDenHaanstat(GEct),pmaxDenHaanstat(GEct)
write(*,"(A,F10.5,F10.5)") "Chg. in Capital Den Haan % Error (avg,max) = ",KavgDenHaanchange,KmaxDenHaanchange
write(*,"(A,F10.5,F10.5)") "Chg. in Price Den Haan % Error (avg,max) = ",pavgDenHaanchange,pmaxDenHaanchange
write(*,*) " "
write(*,*) "Capital Forecast Coefficients (Old, New) "

write(13,*) " "
write(13,"(A,I5)") "GE Iteration ",GEct
write(13,"(A,F10.5,F10.5)") "Fcst Coeff Error (capital, price) = ",kfcsterror,pfcsterror
write(13,"(A,F10.5,F10.5)") "Fcst RMSE Change (capital, price) = ",KRMSEchange,pRMSEchange
write(13,"(A,F10.5,F10.5)") "Fcst R2 Change = ",KR2change,pR2change
write(13,"(A,F10.5,F10.5)") "Capital Den Haan % Error (avg,max) = ",KavgDenHaanstat(GEct),KmaxDenHaanstat(GEct)
write(13,"(A,F10.5,F10.5)") "Price Den Haan % Error (avg,max) = ",pavgDenHaanstat(GEct),pmaxDenHaanstat(GEct)
write(13,"(A,F10.5,F10.5)") "Chg. in Capital Den Haan % Error (avg,max) = ",KavgDenHaanchange,KmaxDenHaanchange
write(13,"(A,F10.5,F10.5)") "Chg. in Price Den Haan % Error (avg,max) = ",pavgDenHaanchange,pmaxDenHaanchange
write(13,*) " "
write(13,*) "Capital Forecast Coefficients (Old, New) "

do act=1,anum
do sct=1,snum
do smin1ct=1,snum
    write(*,"(F10.5,F10.5,F10.5,F10.5)") kfcstcoeff(act,sct,smin1ct,:),kfcstcoeffnew(act,sct,smin1ct,:)
    write(13,"(F10.5,F10.5,F10.5,F10.5)") kfcstcoeff(act,sct,smin1ct,:),kfcstcoeffnew(act,sct,smin1ct,:)
end do !smin1ct
end do !sct
end do !act

write(*,*) " "
write(*,*) "Capital RMSE, R^2"

write(13,*) " "
write(13,*) "Capital RMSE, R^2"

do act=1,anum
do sct=1,snum
do smin1ct=1,snum
    write(*,"(F20.10,A,F20.10)") KRMSEstore(act,sct,smin1ct,GEct),', ',KR2store(act,sct,smin1ct,GEct)
    write(13,"(F20.10,A,F20.10)") KRMSEstore(act,sct,smin1ct,GEct),', ',KR2store(act,sct,smin1ct,GEct)
end do !smin1ct
end do !sct
end do !act


write(*,*) " "
write(*,*) "Price Forecast Coefficients (Old, New) "

write(13,*) " "
write(13,*) "Price Forecast Coefficients (Old, New) "

do act=1,anum
do sct=1,snum
do smin1ct=1,snum
    write(*,"(F10.5,F10.5,F10.5,F10.5)") pfcstcoeff(act,sct,smin1ct,:),pfcstcoeffnew(act,sct,smin1ct,:)
    write(13,"(F10.5,F10.5,F10.5,F10.5)") pfcstcoeff(act,sct,smin1ct,:),pfcstcoeffnew(act,sct,smin1ct,:)
end do !smin1ct
end do !sct
end do !act

write(*,*) " "
write(*,*) "Price RMSE, R^2"

write(13,*) " "
write(13,*) "Price RMSE, R^2"

do act=1,anum
do sct=1,snum
do smin1ct=1,snum
    write(*,"(F20.10,A,F20.10)") pRMSEstore(act,sct,smin1ct,GEct),', ',pR2store(act,sct,smin1ct,GEct)
    write(13,"(F20.10,A,F20.10)") pRMSEstore(act,sct,smin1ct,GEct),', ',pR2store(act,sct,smin1ct,GEct)
end do !smin1ct
end do !sct
end do !act

!!!!!!BLOCK FOR COMPUTING SMM MOMENTS AND BC MOMENTS

write(*,"(A,F15.1,A)") "Started micro tfp calcs at ",omp_get_wtime()-start," seconds."    
write(13,"(A,F15.1,A)") "Started micro tfp calcs at ",omp_get_wtime()-start," seconds."    

!loop over firm years to compute tfp panel
laborannual(:,:) = 0.0
capitalannual(:,:) = 0.0
outputannual(:,:) = 0.0
tfpannual(:,:) = 0.0
investmentannual(:,:) = 0.0
do firmct=1,nfirms
do yrct=1,numyrs
    
    !qtr 1 
    qtrct = (yrct-1)*4+1
    endogct = endogfirms(firmct,qtrct); !indexes (k,l_{-1})
    aval = a0(asimpos(qtrct))
    zval = z0(zfirmspos(firmct,qtrct))
    kval = endog_key(endogct,1)
    polstar = endogfirms(firmct,qtrct+1); !indexes (k',l)
    lval = endog_key(polstar,2) 
    yval = y(zval,aval,kval,lval,alpha,nu)
    laborannual(firmct,yrct) = lval; !insert first quarter labor into appropriate position
    outputannual(firmct,yrct) = outputannual(firmct,yrct) + yval; !increment annual output
    investmentannual(firmct,yrct) = investmentannual(firmct,yrct) + (endog_key(polstar,1) - (1.0 - deltak)*kval)
    
    !qtr 2
    qtrct = (yrct-1)*4+2
    endogct = endogfirms(firmct,qtrct); !indexes (k,l_{-1})
    aval = a0(asimpos(qtrct))
    zval = z0(zfirmspos(firmct,qtrct))
    kval = endog_key(endogct,1)
    polstar = endogfirms(firmct,qtrct+1); !indexes (k',l)
    lval = endog_key(polstar,2) 
    yval = y(zval,aval,kval,lval,alpha,nu)
    outputannual(firmct,yrct) = outputannual(firmct,yrct) + yval; !increment annual output
    investmentannual(firmct,yrct) = investmentannual(firmct,yrct) + (endog_key(polstar,1) - (1.0 - deltak)*kval)
    
    !qtr 3
    qtrct = (yrct-1)*4+3
    endogct = endogfirms(firmct,qtrct); !indexes (k,l_{-1})
    aval = a0(asimpos(qtrct))
    zval = z0(zfirmspos(firmct,qtrct))
    kval = endog_key(endogct,1)
    polstar = endogfirms(firmct,qtrct+1); !indexes (k',l)
    lval = endog_key(polstar,2) 
    yval = y(zval,aval,kval,lval,alpha,nu)
    outputannual(firmct,yrct) = outputannual(firmct,yrct) + yval; !increment annual output
    investmentannual(firmct,yrct) = investmentannual(firmct,yrct) + (endog_key(polstar,1) - (1.0 - deltak)*kval)
    
    !qtr 4
    qtrct = (yrct-1)*4+4
    endogct = endogfirms(firmct,qtrct); !indexes (k,l_{-1})
    aval = a0(asimpos(qtrct))
    zval = z0(zfirmspos(firmct,qtrct))
    kval = endog_key(endogct,1)
    polstar = endogfirms(firmct,qtrct+1); !indexes (k',l)
    lval = endog_key(polstar,2) 
    yval = y(zval,aval,kval,lval,alpha,nu)
    outputannual(firmct,yrct) = outputannual(firmct,yrct) + yval; !increment annual output
    capitalannual(firmct,yrct) = kval; !record 4th-quarter capital as annual capital
    tfpannual(firmct,yrct) = 100.0 * (log(outputannual(firmct,yrct)) - (1.0/3.0)*log(capitalannual(firmct,yrct)) - &
        (2.0/3.0)*log(laborannual(firmct,yrct)))
    investmentannual(firmct,yrct) = investmentannual(firmct,yrct) + (endog_key(polstar,1) - (1.0 - deltak)*kval)
    
end do !yrct
end do !firmct

!write idio TFP series for use by MATLAB
open(8,file=basedir//"\tfpannual.txt")
do firmct=1,nfirms
do yrct=1,numyrs
write(8,*) tfpannual(firmct,yrct)
end do !yrct
end do !firmct
close(8)

!write idio labor series for use by MATLAB
open(8,file=basedir//"\laborannual.txt")
do firmct=1,nfirms
do yrct=1,numyrs
write(8,*) laborannual(firmct,yrct)
end do !yrct
end do !firmct
close(8)

!write idio capital series for use by MATLAB
open(8,file=basedir//"\capitalannual.txt")
do firmct=1,nfirms
do yrct=1,numyrs
write(8,*) capitalannual(firmct,yrct)
end do !yrct
end do !firmct
close(8)

!write idio output series for use by MATLAB
open(8,file=basedir//"\outputannual.txt")
do firmct=1,nfirms
do yrct=1,numyrs
write(8,*) outputannual(firmct,yrct)
end do !yrct
end do !firmct
close(8)

!write idio capital series for use by MATLAB
open(8,file=basedir//"\investmentannual.txt")
do firmct=1,nfirms
do yrct=1,numyrs
write(8,*) investmentannual(firmct,yrct)
end do !yrct
end do !firmct
close(8)

write(*,"(A,F15.1,A)") "Finished micro tfp calcs at ",omp_get_wtime()-start," seconds."    
write(13,"(A,F15.1,A)") "Finished micro tfp calcs at ",omp_get_wtime()-start," seconds."    

write(*,*) " "
write(13,*) " "

write(*,"(A,F15.1,A)") "Started macro tfp calcs at ",omp_get_wtime()-start," seconds."    
write(13,"(A,F15.1,A)") "Started macro tfp calcs at ",omp_get_wtime()-start," seconds."    

!first, what is the aggregate TFP series, via Solow Residual approach?
TFPsim(:) = 0.0
do t=(numdiscard+1),(numper-1)
    TFPsim(t) = log(Ysim(t)) - (1.0/3.0)*log(Ksim(t)) - (2.0/3.0)*log(Lsim(t))
end do !t

!write aggregate TFP series for use by MATLAB
open(8,file=basedir//"\TFPsim.txt")
do t=1,numper
write(8,*) TFPsim(t)
end do !t
close(8)

write(*,"(A,F15.1,A)") "Finished macro tfp calcs at ",omp_get_wtime()-start," seconds."    
write(13,"(A,F15.1,A)") "Finished macro tfp calcs at ",omp_get_wtime()-start," seconds."    

write(*,*) " "
write(13,*) " "

!call MATLAB to do the GARCH(1,1) estimation and compute MACRO & MICRO moments
write(*,*) "Calling MATLAB for GARCH estimation and micro IQR calculations.";
write(13,*) "Calling MATLAB for GARCH estimation and micro IQR calculations.";
call system(matlabstr); !this actually calls MATLAB

!check to see if MATLAB has finished computing the moments, repeatedly
do ct=1,maxwaits

!give MATLAB waitsec seconds to complete the calculation
call sleep(waitsec)

write(*,"(A,I5,A)") "Checking ",ct," times for MATLAB completion."
write(13,"(A,I5,A)") "Checking ",ct," times for MATLAB completion."

!check to see if MATLAB's done
if (file_exists(basedir//"\MICROMOM.txt").and.file_exists(basedir//"\MACROMOM.txt")) exit

if (ct==maxwaits) then

write(*,"(A,I5,A)") "MATLAB failed to complete calculation in ",maxwaits," periods."
write(13,"(A,I5,A)") "MATLAB failed to complete calculation in ",maxwaits," periods."

end if

end do !ct

!now, read the MICRO MOM file from MATLAB
MICROMOM(:) = 0.0
open(8,file=basedir//"\MICROMOM.txt")
do ct=1,4
read(8,*) MICROMOM(ct)
end do !ct
close(8)

!now, read the MACRO MOM file from MATLAB
MACROMOM(:) = 0.0
open(8,file=basedir//"\MACROMOM.txt")
do ct=1,4
read(8,*) MACROMOM(ct)
end do !ct
close(8)

write(*,"(A)") "MATLAB successfully completed calculations."
write(13,"(A)") "MATLAB successfully completed calculations."

write(*,*) " "
write(*,*) "Moments for Macro dTFP GARCH(1,1)-implied Heteroskedasticity" 
write(*,*) "Mean, St Dev, Skewness, Serial Corr"
write(*,"(A,F10.5,F10.5,F10.5,F10.5)") "Data = ",macrodatamom
write(*,"(A,F10.5,F10.5,F10.5,F10.5)") "Model = ",MACROMOM

write(13,*) " "
write(13,*) "Moments for Macro dTFP GARCH(1,1)-implied Heteroskedasticity" 
write(13,*) "Mean, St Dev, Skewness, Serial Corr"
write(13,"(A,F10.5,F10.5,F10.5,F10.5)") "Data = ",macrodatamom
write(13,"(A,F10.5,F10.5,F10.5,F10.5)") "Model = ",MACROMOM

write(*,*) " "
write(13,*) " "

write(*,*) " "
write(*,*) "Moments for Micro TFP resids IQR" 
write(*,*) "Mean, St Dev, Skewness, Serial Corr"
write(*,"(A,F10.5,F10.5,F10.5,F10.5)") "Data = ",microdatamom
write(*,"(A,F10.5,F10.5,F10.5,F10.5)") "Model = ",MICROMOM

write(13,*) " "
write(*,*) "Moments for Micro TFP resids IQR" 
write(13,*) "Mean, St Dev, Skewness, Serial Corr"
write(13,"(A,F10.5,F10.5,F10.5,F10.5)") "Data = ",microdatamom
write(13,"(A,F10.5,F10.5,F10.5,F10.5)") "Model = ",MICROMOM

write(*,*) " "
write(13,*) " "

!!!!!! END BLOCK FOR COMPUTING SMM MOMENTS AND BC MOMENTS

!exit criterion for the GEct looop
exitflag = 0
!coefficient convergence
if (GEerrorswitch==1) then
    if (pfcsterror<fcsterrortol.and.kfcsterror<fcsterrortol) exitflag = 1
!RMSE convergence    
else if (GEerrorswitch==2) then
    if (KRMSEchange<RMSEchangetol.and.pRMSEchange<RMSEchangetol) exitflag = 1
!R2 convergence    
else if (GEerrorswitch==3) then
    if (KR2change<R2changetol.and.pR2change<R2changetol) exitflag = 1
!max Den Haan stat convergence    
else if (GEerrorswitch==4) then
    if (KmaxDenHaanchange<maxDenHaanchangetol.and.pmaxDenHaanchange<maxDenHaanchangetol) exitflag = 1    
!avg Den Haan stat convergence    
else if (GEerrorswitch==5) then
    if (KavgDenHaanchange<avgDenHaanchangetol.and.pavgDenHaanchange<avgDenHaanchangetol) exitflag = 1    
end if

!this actually exits the GE loop if you've converged on the desired metric
if (exitflag==1) exit

!update the forecast rule coefficients with gain - i.e. dampened fixed point iteration
kfcstcoeff = kfcstcoeff + GEupdate * (kfcstcoeffnew - kfcstcoeff)
pfcstcoeff = pfcstcoeff + GEupdate * (pfcstcoeffnew - pfcstcoeff)

!if the forecast rule has not converged, need to update forecast matrices used
!in the VFI for the next GE iteration
do act=1,anum
do sct=1,snum
do smin1ct=1,snum

    !insert the forecasts into the fcst array
    do fcstct=1,numfcst
        kbarval = kbar0(fcstct); !what is the value of aggregate capital?
        kbarfcstval = exp( kfcstcoeff(act,sct,smin1ct,1) + kfcstcoeff(act,sct,smin1ct,2) * log( kbarval ) ); !implied agg cap fcst
        pfcstval = exp( pfcstcoeff(act,sct,smin1ct,1) + pfcstcoeff(act,sct,smin1ct,2) * log( kbarval ) ); !implied price fcst
        
        !adjust the agg capital forecast to the boundaries of the grid
        kbarfcstval = minval( (/ maxval( (/ kbarmin , kbarfcstval /) ) , kbarmax /) )
        
        !find the interval of kbar0 in which kbarfcst lands, as well as the lin. interp. weight
        ind = fcstct; call hunt(kbar0,numfcst,kbarfcstval,ind)
        weight = ( kbarfcstval - kbar0(ind) ) / ( kbar0(ind+1) - kbar0(ind) )
        
            !got to make sure that out of bounds entries are handled correctly
        if (ind>=1.and.ind<kbarnum) then
            kbarfcstinds(act,sct,smin1ct,fcstct) = ind; kbarfcstweights(act,sct,smin1ct,fcstct) = weight
        else if (ind==0) then
            kbarfcstinds(act,sct,smin1ct,fcstct) = 1; kbarfcstweights(act,sct,smin1ct,fcstct) = 0.0
        else if (ind==kbarnum) then
            kbarfcstinds(act,sct,smin1ct,fcstct) = kbarnum-1; kbarfcstweights(act,sct,smin1ct,fcstct) = 1.0
        end if        
        
        !insertion into arrays, including insertion of the implied wage
        kbarfcstmat(act,sct,smin1ct,fcstct) = kbarfcstval
        pfcstmat(act,sct,smin1ct,fcstct) = pfcstval
        wfcstmat(act,sct,smin1ct,fcstct) = theta / pfcstval
        
    end do !fcstct
end do !smin1ct
end do !sct
end do !act

!!!!END THE FCST RULE UPDATE BLOCK

end do; !GEct - THIS ENDS THE GE FCST RULE ITERATING

!call the clock to get the end time and closes log file
finish = omp_get_wtime()
write(*,"(A,F15.1,A)") "Finished completely in ",finish-start," seconds."
write(13,"(A,F15.1,A)") "Finished completely in ",finish-start," seconds."
close(13)

!actually construct the SMM objective
RUBC_func = 0.0
do ct=1,4
RUBC_func = RUBC_func + ((macrodatamom(ct) - MACROMOM(ct))/macrodatamom(ct))**2.0
RUBC_func = RUBC_func + ((microdatamom(ct) - MICROMOM(ct))/microdatamom(ct))**2.0
end do !ct

!now, penalize grid bunching
!if (maxval((/ TOPkstat,BOTkstat,TOPlstat,BOTlstat/))>0.025) then
if ((TOPkstat+BOTkstat+TOPlstat+BOTlstat)>0.025) then
    RUBC_func = 1000.0
end if

open(8,file=basedir//"\SMMobjectivestore.txt",position='append')
write(8,*) RUBC_func
close(8)

open(8,file=basedir//"\SMMparamstore.txt",position='append')
write(8,*) x(1),x(2),x(3),x(4),x(5),x(6)
close(8)

open(8,file=basedir//"\MACROMOMstore.txt",position='append')
write(8,*) MACROMOM(1),MACROMOM(2),MACROMOM(3),MACROMOM(4)
close(8)

open(8,file=basedir//"\MICROMOMstore.txt",position='append')
write(8,*) MICROMOM(1),MICROMOM(2),MICROMOM(3),MICROMOM(4)
close(8)

open(8,file=basedir//"\GRIDBUNCHstore.txt",position='append')
write(8,*) TOPkstat,BOTkstat,TOPlstat,BOTlstat
close(8)

open(8,file=basedir//"\GEctstore.txt",position='append')
write(8,*) GEct-1
close(8)

open(8,file=basedir//"\TIMEstore.txt",position='append')
write(8,*) finish-start
close(8)

open(8,file=basedir//"\MAXDHchangestore.txt",position='append')
write(8,*) pmaxDenHaanchange,KmaxDenHaanchange
close(8)

write(*,*) "OBJECTIVE FUNCTION = ",RUBC_func
write(*,*) "##########################################"

end function RUBC_func

end program RUBC_wrapper